"""Define version and import high-level objects."""
__VERSION_MAJOR__ = '1'
__VERSION_MINOR__ = '0'
__VERSION_BUILD__ = '3'

__DEBUG = False
if __DEBUG:
    import datetime as dt
    __VERSION_TAGS__ = f"d{dt.datetime.now().strftime('%Y%m%d_%H%M')}"
else:
    __VERSION_TAGS__ = ''

__VERSION__ = f'{__VERSION_MAJOR__}.{__VERSION_MINOR__}.{__VERSION_BUILD__}{__VERSION_TAGS__}'

__DATA_VERSION__ = '1.1'

try:
    from audata.file import File
except ImportError:
    print("Unable to import audata.file.File. If this occurs during initial setup, you may ignore this.")


"""Tool for automagically converting CSV files to audata files."""
import os
import argparse
from glob import glob
import datetime as dt
from typing import Optional

from dateutil.parser import parse, ParserError
import pandas as pd
import h5py as h5

import audata


def _mkfn(path: str) -> str:
    """Convert directory or CSV filename to h5 filename."""
    if os.path.isdir(path):
        return os.path.basename(path) + '.h5'
    else:
        return os.path.splitext(os.path.basename(path))[0] + '.h5'


def _addcsv(au_file: audata.File, name: str, path: str):
    """Add CSV file to the audata file."""
    data = pd.read_csv(path)
    for col in data.columns:
        # Does it have a name that can be interpreted as a "numeric" time?
        is_datetime = col.lower() in ('time', 'timestamp', 'index')
        is_str = data[col].dtype in (str, h5.string_dtype())

        if is_datetime or is_str:
            # Can we convert to date?
            if is_str:
                try:
                    _ = parse(data[col].values[0])
                    is_datetime = True
                except ParserError:
                    pass

            if is_datetime:
                if data[col].dtype in (str, h5.string_dtype()):
                    print('    Parsing time string column: {}'.format(col))
                    data[col] = data[col].apply(parse)
                else:
                    print('    Parsing time offset column: {}'.format(col))
                    data[col] = au_file.time_reference + data[col] * dt.timedelta(seconds=1)

            else:
                arity = len(data[col].unique())
                pct = float(arity) / len(data) * 100
                if pct > 10:
                    print('    Found string column: {} (arity {}%)'.format(
                        col, round(pct)))
                else:
                    print('    Parsing categorical column: {} (arity {} / {}%)'.
                          format(col, arity, round(pct)))
                    data[col] = pd.Series(data[col], dtype='category')
        else:
            print('    Found {} column: {}'.format(data[col].dtype, col))
    au_file[name] = data


def _walk(au_file: audata.File, path: str, prefix: Optional[str] = None):
    """Recurse a directory in search of CSV files to add."""
    if path is None or len(path) == 0 or path[0] == '.':
        return
    elif os.path.isdir(path):
        prefix = '' if prefix is None else os.path.basename(path)
        for next_path in glob('{}/*'.format(path)):
            _walk(au_file, next_path, prefix)
    elif os.path.splitext(path)[1].lower() == '.csv':
        if prefix is None:
            prefix = ''
        name = '{}{}{}'.format(prefix, '/' if prefix != '' else '',
                               os.path.splitext(os.path.basename(path))[0])
        print('  adding {}'.format(name))
        _addcsv(au_file, name, path)
    else:
        return


def main():
    """
    Converts CSV files or nested directories containing multiple CSVs to an audata HDF5 file.

    This tool takes in the path to either a single CSV file or a directory to be recursed for all
    CSV files. For each, it attempts to intelligently convert each CSV into an `audata` dataset
    and store the output to a file (with the same name as the CSV or directory, ending in '.h5').

    The tool will infer a datetime column if it is named time, timestamp, or index, or if it is a
    string column whose first element can be parsed as a datetime object using
    `dateutils.parser.parse`. This parsing does rely on the first row being valid.

    Other string columns will be treated either as factor/categorical columns if the arity (unique
    values) is less than 10% of the total number of rows, otherwise a string column.

    Args:
        path (str): Path to a CSV file or directory to be recursively scanned for CSVs.
    """
    parser = argparse.ArgumentParser(
        description=
        'Converts CSV files or nested directories containing multiple CSVs to an audata HDF5 file.'
    )
    parser.add_argument(
        'path',
        type=str,
        help=
        'Path to a CSV file or directory to be recursively scanned for CSVs.')
    args = parser.parse_args()

    filename = _mkfn(args.path)
    print('Creating {}'.format(filename))
    with audata.File.new(filename, overwrite=True) as au_file:
        _walk(au_file, args.path)
        print(au_file)


if __name__ == '__main__':
    main()



"""Wrapper for Group types."""
from typing import Dict, List, Iterable, Tuple, Union

import h5py as h5

from audata.element import Element
from audata.dataset import Dataset


class Group(Element):
    """
    Group element. Acts largely like a container for datasets. Generally should not be
    instantiated directly.
    """

    def __init__(self, au_parent: Element, name: str = ''):

        if not isinstance(au_parent, (Element, h5.File)):
            raise Exception(f'Invalid parent: {type(au_parent)}')

        if isinstance(au_parent, Element):
            parent = au_parent.hdf
            if not isinstance(parent, h5.Group):
                raise Exception(f'Invalid parent: {type(parent)}')

            if name == '':
                target = parent
            else:
                if name not in parent:
                    raise Exception(
                        f'Path {name} was not found in {parent.file.filename}:{parent.name}'
                    )

                target = parent[name]
            if not isinstance(target, h5.Group):
                raise Exception(
                    f'Path "{name}" is not a group in {parent.file.filename}:{parent.name}'
                )

        super().__init__(au_parent, name)

    def list(self) -> Dict[str, List[str]]:
        """List all child attributes, groups, and datasets."""
        attrs = list(self.hdf.attrs)
        others = list(self.hdf)
        groups = [g for g in others if isinstance(self.hdf[g], h5.Group)]
        datasets = [d for d in others if isinstance(self.hdf[d], h5.Dataset)]
        return {'attributes': attrs, 'groups': groups, 'datasets': datasets}

    def recurse(self) -> Iterable[Tuple[Element, str]]:
        """
        Recursively find all datasets. Groups and datasets prefixed with a period
        are ignored.

        Returns:
            Iterable (generator) of tuples of (object: Element, name: str).
        """
        names = [n for n in list(self.hdf) if not n.startswith('.')]
        for name in names:
            elem = self.__getitem__(name)
            if isinstance(elem, Group):
                for stuple in elem.recurse():
                    yield stuple
            else:
                yield (elem, elem.name)

    def __repr__(self):
        lines = []
        elems = self.list()
        if self.name in ('', '/'):
            lines.append('<ROOT>')
        else:
            lines.append(self.name)

        for attr in elems['attributes']:
            lines.append(f'  [A] {attr}')
        for group in elems['groups']:
            lines.append(f'  [G] {group}')
        for dataset in elems['datasets']:
            lines.append(f'  [D] {dataset}')
        return '\n'.join(lines)

    def __str__(self):
        return self.__repr__()

    def __getitem__(self, key: str) -> Union['Dataset', 'Group', None]:
        if self.hdf is None:
            raise Exception('No group opened.')

        if key == '':
            return Group(self, key)

        if key in self.hdf:
            if isinstance(self.hdf[key], h5.Dataset):
                return Dataset(self, key)
            elif isinstance(self.hdf[key], h5.Group):
                return Group(self, key)
            else:
                raise Exception('Unsure how to handle class: {}'.format(
                    type(self.hdf[key])))
        else:
            return None

    def __setitem__(self,
                    key: str,
                    value: Union[h5.Dataset, 'np.ndarray', 'np.recarray',
                                 'pd.DataFrame', None],
                    overwrite: bool = True,
                    **kwargs):

        if self.hdf is None:
            raise Exception('No group opened.')

        if value is None:
            if key in self.hdf:
                del self.hdf[key]
        else:
            Dataset.new(self, key, value, overwrite=overwrite, **kwargs)

    def __contains__(self, key: str) -> bool:
        return self.hdf.__contains__(key)

    def new_dataset(
            self, name: str,
            value: Union[h5.
                         Dataset, 'np.ndarray', 'np.recarray', 'pd.DataFrame'],
            **kwargs):
        """Create a new dataset."""

        self.__setitem__(name, value, **kwargs)



"""Class and related functionality for projects."""
import logging
import pandas as pd
import traceback

from pathlib import Path
from sqlalchemy import and_
from sqlalchemy.orm import contains_eager, joinedload
from typing import AnyStr, List, Dict, Optional, Union

from . import models
from .patternset import PatternSet
from .config import config
from .file import File
from .shared import annotationDataFrame, annotationOrPatternOutput, patternDataFrame

class Project:
    """Represents an auviewer project."""

    def __init__(self, projectModel, processNewFiles=True):
        """The project name should also be the directory name in the projects directory."""

        # Set id, name, and relevant paths
        self.id = projectModel.id
        self.name = projectModel.name
        self.projDirPathObj = Path(projectModel.path)
        self.originalsDirPathObj = self.projDirPathObj / 'originals'
        self.processedDirPathObj = self.projDirPathObj / 'processed'

        # Hold a reference to the db model for future use
        self.model = projectModel

        # Load interface templates
        self.interfaceTemplates = "{}"
        p = self.projDirPathObj / 'templates' / 'interface_templates.json'
        if p.is_file():
            with p.open() as f:
                self.interfaceTemplates = f.read()

        # Load project template
        self.projectTemplate = "{}"
        p = self.projDirPathObj / 'templates' / 'project_template.json'
        if p.is_file():
            with p.open() as f:
                self.projectTemplate = f.read()

        # Holds references to the files that belong to the project
        self.files = []

        # Holds references to the pattern sets that belong to the project,
        # indexed by pattern set ID.
        self.patternsets = {}

        # Load pattern sets
        self.loadPatternSets()

        # Load project files
        self.loadProjectFiles(processNewFiles)

    def __del__(self):
        """Cleanup"""
        try:
            self.observer.join()
        except:
            pass

    def createPatternSet(self, name: str, description=None, showByDefault: bool = False) -> PatternSet:
        """
        Create and return a new pattern set.
        :return: a new PatternSet instance
        """

        # Create pattern set in the database
        patternSetModel = models.PatternSet(project_id=self.id, name=name, description=description, show_by_default=showByDefault)
        models.db.session.add(patternSetModel)
        models.db.session.commit()

        # Instantiate PatternSet and add to the project's pattern sets
        ps = PatternSet(self, patternSetModel)
        self.patternsets[ps.id] = ps

        # Return the pattern set
        return ps

    def detectPatterns(self, type, series, thresholdlow, thresholdhigh, duration, persistence, maxgap, expected_frequency=0, min_density=0):
        """
        Run pattern detection on all files, and return a DataFrame of results.
        This DataFrame, or a subset thereof, can be passed into PatternSet.addPatterns() if desired.
        """
        patterns = [[f.id, f.name, series, pattern[0], pattern[1], None, None] for f in self.files for pattern in f.detectPatterns(type, series, thresholdlow, thresholdhigh, duration, persistence, maxgap, expected_frequency=expected_frequency, min_density=min_density)]
        pdf = pd.DataFrame(patterns, columns=['file_id', 'filename', 'series', 'left', 'right', 'top', 'bottom'])
        pdf['label'] = ''
        return pdf

    def getAnnotations(
            self,
            annotation_id: Union[int, List[int], None] = None,
            file_id: Union[int, List[int], None] = None,
            pattern_id: Union[int, List[int], None] = None,
            pattern_set_id: Union[int, List[int], None]=None,
            series: Union[AnyStr, List[AnyStr], None]=None,
            user_id: Union[int, List[int], None] = None) -> pd.DataFrame:
        """
        Returns a dataframe of annotations for this project, optionally filtered.
        """

        # Prepare input
        if not isinstance(annotation_id, List) and annotation_id is not None:
            annotation_id = [annotation_id]
        if not isinstance(file_id, List) and file_id is not None:
            file_id = [file_id]
        if not isinstance(pattern_id, List) and pattern_id is not None:
            pattern_id = [pattern_id]
        if not isinstance(pattern_set_id, List) and pattern_set_id is not None:
            pattern_set_id = [pattern_set_id]
        if not isinstance(series, List) and series is not None:
            series = [series]
        if not isinstance(user_id, List) and user_id is not None:
            user_id = [user_id]

        # Query
        q = models.Annotation.query.options(joinedload('user'))

        # Filter query as necessary
        if annotation_id is not None:
            q = q.filter(models.Annotation.id.in_(annotation_id))
        if file_id is not None:
            q = q.filter(models.Annotation.file_id.in_(file_id))
        if pattern_id is not None:
            q = q.filter(models.Annotation.pattern_id.in_(pattern_id))
        if pattern_set_id is not None:
            q = q.filter(models.Annotation.pattern_set_id.in_(pattern_set_id))
        if series is not None:
            q = q.filter(models.Annotation.series.in_(series))
        if user_id is not None:
            q = q.filter(models.Annotation.user_id.in_(user_id))

        # Return the dataframe
        return annotationDataFrame(q.filter(models.Annotation.project_id == self.id).all())

    def getAnnotationsOutput(self, user_id: int):
        """Returns a list of user's annotations for all files in the project"""
        return [[a.id, a.file_id, Path(a.file.path).name, a.series, a.left, a.right, a.top, a.bottom, a.label, a.pattern_id] for a in models.Annotation.query.filter_by(user_id=user_id, project_id=self.id).all()]

    def getFile(self, id):
        """Returns the file with matching ID or None."""
        for f in self.files:
            if f.id == id:
                return f
        return None

    def getInitialPayload(self, user_id):
        """Returns initial project payload data"""

        print("Assembling initial project payload output for project", self.name)

        return {

            # Project data
            'project_id': self.id,
            'project_name': self.name,
            'project_assignments': [{
                'id': ps.id,
                'name': ps.name,
                'description': ps.description,
                'patterns': [
                    #annotationOrPatternOutput(p) for p in ps.patterns
                    annotationOrPatternOutput(p, p.annotations[0] if len(p.annotations)>0 else None) for p in models.db.session.query(models.Pattern).filter_by(pattern_set_id=ps.id).outerjoin(models.Annotation, and_(models.Annotation.pattern_id==models.Pattern.id, models.Annotation.user_id==user_id)).options(contains_eager('annotations')).all()
                ]
            } for ps in models.PatternSet.query.filter(models.PatternSet.users.any(id=user_id), models.PatternSet.project_id==self.id).all()],
            'project_files': [[f.id, f.origFilePathObj.name] for f in self.files],

            # Template data
            'builtin_default_interface_templates': config['builtinDefaultInterfaceTemplates'],
            'builtin_default_project_template': config['builtinDefaultProjectTemplate'],
            'global_default_interface_templates': config['globalDefaultInterfaceTemplates'],
            'global_default_project_template': config['globalDefaultProjectTemplate'],
            'interface_templates': self.interfaceTemplates,
            'project_template': self.projectTemplate,

        }

    def getPatterns(
            self,
            file_id: Union[int, List[int], None] = None,
            pattern_id: Union[int, List[int], None] = None,
            pattern_set_id: Union[int, List[int], None] = None,
            series: Union[AnyStr, List[AnyStr], None] = None,
            user_id: Union[int, List[int], None] = None) -> pd.DataFrame:
        """Returns a dataframe of patterns for this project, optionally filtered."""

        # Prepare input
        if not isinstance(file_id, List) and file_id is not None:
            file_id = [file_id]
        if not isinstance(pattern_id, List) and pattern_id is not None:
            pattern_id = [pattern_id]
        if not isinstance(pattern_set_id, List) and pattern_set_id is not None:
            pattern_set_id = [pattern_set_id]
        if not isinstance(series, List) and series is not None:
            series = [series]
        if not isinstance(user_id, List) and user_id is not None:
            user_id = [user_id]

        # Query
        q = models.Pattern.query

        # Filter query as necessary
        if file_id is not None:
            q = q.filter(models.Pattern.file_id.in_(file_id))
        if pattern_id is not None:
            q = q.filter(models.Pattern.pattern_id.in_(pattern_id))
        if pattern_set_id is not None:
            q = q.filter(models.Pattern.pattern_set_id.in_(pattern_set_id))
        if series is not None:
            q = q.filter(models.Pattern.series.in_(series))
        if user_id is not None:
            q = q.filter(models.Pattern.user_id.in_(user_id))

        # Return the dataframe
        return patternDataFrame(q.filter(models.Pattern.project_id == self.id).all())


    def getPatternSet(self, id) -> Optional[PatternSet]:
        """
        Get project's pattern set by ID.
        :return: the PatternSet instance belonging to the id, or None if not found
        """
        return self.patternsets[id] if id in self.patternsets else None

    def getPatternSets(self) -> Dict[int, PatternSet]:
        """
        Get project's pattern sets.
        :return: a dict of the project's PatternSet instances, indexed by id
        """
        return self.patternsets.copy()

    def getTotalPatternCount(self) -> int:
        """
        Get total count of patterns in all the project's pattern sets
        :return: number of patterns
        """
        return sum([ps.count for ps in self.patternsets.values()])

    def listPatternSets(self) -> List[List[str]]:
        """
        Returns list of pattern sets (ID, names).
        :return: list of l:return:
        """
        return [[ps.id, ps.name] for ps in self.patternsets.values()]

    def loadPatternSets(self) -> None:
        """Load or reload the project's pattern sets."""

        # Reset pattern sets container
        self.patternsets = {}

        # Load pattern sets
        for psm in models.PatternSet.query.filter_by(project_id=self.id).all():
            self.patternsets[psm.id] = PatternSet(self, psm)

    def loadProjectFiles(self, processNewFiles=True):
        """Load or reload files belonging to the project, and process new files if desired."""

        # Reset files list
        self.files = []

        # Will hold all project files that exist in the database (in order to
        # detect new files to process).
        existingFilePathObjs = []

        # Get all of the project's files listed in the database
        fileDBModels = models.File.query.filter_by(project_id=self.id).all()

        # For each project file in the database...
        for fileDBModel in fileDBModels:

            # Verify the original file exists on the file system
            origFilePathObj = Path(fileDBModel.path)
            existingFilePathObjs.append(origFilePathObj)
            if not origFilePathObj.exists():
                logging.error(
                    f"File ID {fileDBModel.id} in the database is missing the original file on the file system at {fileDBModel.path}")
                continue
                # TODO(gus): Do something else with this? Like set error state in the database entry and display in GUI?

            # Verify the processed file exists on the file system
            procFilePathObj = self.processedDirPathObj / (origFilePathObj.stem + '_processed.h5')
            if not procFilePathObj.exists():
                logging.error(
                    f"File ID {fileDBModel.id} in the database is missing the processed file on the file system at {procFilePathObj}")
                continue
                # TODO(gus): Do something else with this? Like set error state in the database entry and display in GUI?

            # Instantiate the file class, and attach to this project instance
            self.files.append(File(self, fileDBModel.id, origFilePathObj, procFilePathObj, processNewFiles))

        # If processNewFiles is true, then go through and process new files
        if processNewFiles:

            # For each new project file which does not exist in the database...
            for newOrigFilePathObj in [p for p in self.originalsDirPathObj.iterdir() if p.is_file() and p.suffix == '.h5' and not any(map(lambda existingFilePathObj: p.samefile(existingFilePathObj), existingFilePathObjs))]:

                # Establish the path of the new processed file
                newProcFilePathObj = self.processedDirPathObj / (newOrigFilePathObj.stem + '_processed.h5')

                # Instantiate the file class with an id of -1, and attach to
                # this project instance.
                try:
                    newFileClassInstance = File(self, -1, newOrigFilePathObj, newProcFilePathObj, processNewFiles)
                except Exception as e:
                    logging.error(f"New file {newOrigFilePathObj} could not be processed.\n{e}\n{traceback.format_exc()}")
                    continue

                # Now that the processing has completed (if not, an exception
                # would have been raised), add the file to the database and
                # update the file class instance ID.
                newFileDBEntry = models.File(project_id=self.id, path=str(newOrigFilePathObj))
                models.db.session.add(newFileDBEntry)
                models.db.session.commit()

                # Update the file class instance ID, and add it to the files
                # list for this project.
                newFileClassInstance.id = newFileDBEntry.id
                self.files.append(newFileClassInstance)

    def setName(self, name):
        """Rename the project."""
        self.model.name = name
        models.db.session.commit()
        self.name = name



"""Helper utilities."""
import datetime as dt
import json
from typing import Optional, Dict, Any, AbstractSet, Tuple, Union

import pandas as pd
import numpy as np
import h5py as h5
import jsbeautifier as jsb


def df_from_audata(rec,
                   columns: Dict[str, Any],
                   time_ref: Optional[dt.datetime] = None,
                   datetimes: bool = True) -> pd.DataFrame:
    """Create a pandas DataFrame from an audata Dataset."""

    data = pd.DataFrame(data=rec)
    for col in columns:
        col_meta = columns[col]
        if col_meta['type'] == 'factor':
            data[col] = pd.Categorical.from_codes(data[col].values, col_meta['levels'])
        elif col_meta['type'] == 'time':
            if time_ref is None:
                raise Exception('Cannot read timestamps without reference!')
            if datetimes:
                data[col] = time_ref + data[col].values * dt.timedelta(seconds=1)
            else:
                data[col] += time_ref.timestamp()
        elif col_meta['type'] == 'timedelta':
            data[col] = data[col].values * dt.timedelta(seconds=1)
    return data


def audata_from_df(data: pd.DataFrame,
                   time_ref: Optional[dt.datetime] = None,
                   time_cols: Optional[AbstractSet[str]] = None,
                   timedelta_cols: Optional[AbstractSet[str]] = None
                  ) -> Tuple[Dict[str, Any], np.recarray]:
    """Create the recarray and meta from a DataFrame to be stored to the audata file."""

    if time_cols is None:
        time_cols = set({})
    if timedelta_cols is None:
        time_cols = set({})
    cols = list(data)
    columns = {}
    dtype_map = {}
    for col in cols:
        col_meta = {}
        col_dtype = data[col].dtype

        if col_dtype == h5.string_dtype():
            # String d-type has to be set explicitely or HDF5 won't accept it. Default to
            # variable-length strings, although in the future we might want to support fixed-
            # length strings which can be compressed (vlen strings are NOT compressed).
            dtype_map[col] = h5.string_dtype()
            col_meta['type'] = 'string'
        elif col_dtype.name == 'category':
            col_meta['type'] = 'factor'
            col_meta['levels'] = list(data[col].cat.categories)
            col_meta['ordered'] = data[col].cat.ordered
            data[col] = data[col].cat.codes
        elif col_dtype.kind in ['i', 'u']:
            col_meta['type'] = 'integer'
            col_meta['signed'] = col_dtype.kind == 'i'
        elif col_dtype.kind == 'M':
            if time_ref is None:
                raise (Exception('Cannot convert timestamps without time reference!'))

            col_meta['type'] = 'time'
            data[col] = (data[col].dt.tz_convert(time_ref.tzinfo) -
                         time_ref).dt.total_seconds()
        elif col_dtype.kind == 'm':
            col_meta['type'] = 'timedelta'
            data[col] = data[col].dt.total_seconds()
        elif col in time_cols:
            # Assume offset from reference in appropriate units.
            col_meta['type'] = 'time'
            dtype_map[col] = 'f8'
        elif col in timedelta_cols:
            # Assume delta in appropriate units.
            col_meta['type'] = 'timedelta'
            dtype_map[col] = 'f8'
        else:
            typenames = {'b': 'boolean', 'f': 'real', 'c': 'complex'}
            col_meta['type'] = typenames[col_dtype.kind]
        columns[col] = col_meta

    meta = {'columns': columns}
    rec = data.to_records(index=False, column_dtypes=dtype_map)
    return meta, rec


def audata_from_arr(arr: Union[np.ndarray, np.recarray],
                    time_ref: Optional[dt.datetime] = None,
                    time_cols: Optional[AbstractSet[str]] = None,
                    timedelta_cols: Optional[AbstractSet[str]] = None
                   ) -> Tuple[Dict[str, Any], np.recarray]:
    """Create the recarray and meta from a recarray or ndarray to be stored to the audata file."""

    # Issue 4: Factor types are not supported in this mode. Also not supported: non-string objects.
    if time_cols is None:
        time_cols = set({})
    if timedelta_cols is None:
        time_cols = set({})
    cols = arr.dtype.names
    columns = {}
    for col in cols:
        col_meta = {}
        col_dtype = arr.dtype[col]

        if col_dtype.kind == 'M':
            if time_ref is None:
                raise (Exception('Cannot convert timestamps without time reference!'))

            # Note: Since numpy datetime64 types are not timezone aware, we must
            # assume the datetimes use the same timezone.
            col_meta['type'] = 'time'
            arr[col] = (arr[col] - np.datetime64(time_ref.replace(tzinfo=None))) / np.timedelta64(1, 's')
        elif col_dtype.kind == 'm':
            col_meta['type'] = 'timedelta'
            arr[col] = arr[col] / np.timedelta64(1, 's')
        elif col_dtype == h5.string_dtype():
            if h5.check_vlen_dtype(col_dtype) != str:
                arr[col] = arr[col].astype(h5.string_dtype())
            col_meta['type'] = 'string'
        elif col in time_cols:
            # Assume offset from reference in appropriate units.
            col_meta['type'] = 'time'
            if col_dtype.type != 'f':
                arr[col] = arr[col].astype('f8')
        elif col in timedelta_cols:
            # Assume delta in appropriate units.
            col_meta['type'] = 'timedelta'
            if col_dtype.type != 'f':
                arr[col] = arr[col].astype('f8')
        elif col_dtype.kind in ['i', 'u']:
            col_meta['type'] = 'integer'
            col_meta['signed'] = col_dtype.kind == 'i'
        else:
            typenames = {'b': 'boolean', 'f': 'real', 'c': 'complex'}
            col_meta['type'] = typenames[col_dtype.kind]
        columns[col] = col_meta

    meta = {'columns': columns}
    return meta, arr

# Makes a best effort to determine column type based only on dtype. Returns a
# dict containing keys 'type' and, if applicable, 'signed'.
# TODO(gus): At somep point, it would be nice to merge this with how
# audata_from_df() and audata_from_arr() determine column type.
def get_coltype_from_dtype_only(dtype) -> Dict[str, Any]:

    if dtype.kind == 'M':
        return { 'type': 'time' }
    elif dtype.kind == 'm':
        return { 'type': 'timedelta' }
    elif dtype == h5.string_dtype():
        return { 'type': 'string' }
    elif dtype.kind == 'i':
        return {
            'type': 'integer',
            'signed': False
        }
    elif dtype.kind == 'u':
        return {
            'type': 'integer',
            'signed': True
        }
    elif dtype.kind == 'b':
        return { 'type': 'boolean' }
    elif dtype.kind == 'f':
        return {'type': 'real' }
    elif dtype.kind == 'c':
        return {'type': 'complex' }
    else:
        raise Exception('Unrecognized column type.')

def json2dict(json_str: str) -> Dict[str, Any]:
    """Convert JSON string to python dictionary."""
    return json.loads(json_str)


def dict2json(json_dict: Dict[str, Any], beautify: bool = True) -> str:
    """Convert JSON-compatible python dictionary to beautified JSON string."""
    if not isinstance(json_dict, dict):
        raise Exception(f'Expecting dictionary, found {type(json_dict)}')

    if beautify:
        return jsb.beautify(json.dumps(json_dict))
    else:
        return json.dumps(json_dict)



from pathlib import Path
from sqlalchemy import distinct, or_
import logging
import time
import traceback
import pandas as pd

import audata

from . import models
from .cylib import generateThresholdAlerts
from .series import Series, simpleSeriesName
from .shared import annotationOrPatternOutput

# File represents a single patient data file. File may operate in file- or
# realtime-mode. In file-mode, all data is written to & read from a file. In
# realtime-mode, no file is dealt with and instead everything is kept in memory.
# If no filename parameter is passed to the constructor, File will operate in
# realtime-mode.
class File:

    def __init__(self, projparent, id, origFilePathObj, procFilePathObj, processNewFiles=True):

        logging.info(f"\n------------------------------------------------\nACCESSING FILE: (ID {id}) {origFilePathObj}\n------------------------------------------------\n")

        # Holds reference to the project parent
        self.projparent = projparent

        # File ID & paths
        self.id = id
        self.origFilePathObj = origFilePathObj
        self.procFilePathObj = procFilePathObj

        # Filename
        self.name = Path(self.origFilePathObj).name

        # Will hold the data series from the file
        self.series = []

        if self.mode() == 'file':

            # Open the original file
            self.f = audata.File.open(str(self.origFilePathObj), return_datetimes=False)

            # Load the processed file if it exists
            if self.procFilePathObj.exists():
                logging.info(f"Opening processed file {self.procFilePathObj}.")
                self.pf = audata.File.open(str(self.procFilePathObj), return_datetimes=False)

            # Load series data into memory
            self.load()

            # If the processed file does not exist and we're supposed to process
            # new file data, process it.
            if not self.procFilePathObj.exists() and processNewFiles:
                logging.info(f"Generating processed file for {self.origFilePathObj}")
                self.process()

            # If the processed file still does not exist, raise an exception
            if not self.procFilePathObj.exists():
                raise Exception(f"File {self.origFilePathObj} has not been processed and therefore cannot be loaded.")

    def __del__(self):

        logging.info("Cleaning up closed file.")

        # Close the original file
        try:
            self.f.close()
        except:
            pass

        # Close the processed file
        try:
            self.pf.close()
        except:
            pass

    # Takes new data to add to one or more data series for the file (currently
    # works only in realtime-mode). The new data is assumed to occur after any
    # existing data. The parameter, seriesData, should be a dict of dicts of
    # lists as follows:
    #
    #   {
    #     'seriesName': {
    #       'times': [ t1, t2, ... , tn ],
    #       'values': [ v1, v2, ... , vn ]
    #     }, ...
    #   }
    #
    # Returns updated file data for transmission to realtime subscribers.
    def addSeriesData(self, seriesData):

        logging.info(f"Adding series data: {seriesData}")

        if self.mode() != 'realtime':
            raise Exception('Can only addSeriesData() while in mem-mode.')

        if not isinstance(seriesData, dict):
            raise Exception('Invalid seriesData parameter received for addSeriesData().')

        # This will hold the update data to return
        update = {
            'series': {}
        }

        for s in seriesData:

            if not isinstance(seriesData[s], dict):
                raise Exception('Invalid seriesData[s] parameter received for addSeriesData().')

            # Retrieve or create the series
            series = self.getSeriesOrCreate(s)

            # If that failed, raise an exception
            if series is None:
                raise Exception('Attempting to create or retrieve series failed:', s)

            # Add the new series data
            series.addData(seriesData[s])

            nones = [None] * len(seriesData[s]['times'])

            update['series'][s] = {
                "id": s,
                "labels": ['Date/Offset', 'Min', 'Max', simpleSeriesName(s)],
                "data": [list(i) for i in zip(seriesData[s]['times'], nones, nones, seriesData[s]['values'])],
                "output_type": 'real'
            }

        logging.info(f"Adding series data completed. Returning {update}")

        return update

    # Create an annotation for the file
    def createAnnotation(self, user_id, left=None, right=None, top=None, bottom=None, seriesID='', label='', pattern_id=None):

        # If the pattern_id is set, determine the pattern_set_id.
        pattern_set_id = None;
        if pattern_id:
            pattern_set_id = models.Pattern.query.filter_by(id=pattern_id).first().pattern_set_id

        newann = models.Annotation(
            user_id=user_id,
            project_id=self.projparent.id,
            file_id=self.id,
            series=seriesID,
            left=left,
            right=right,
            top=top,
            bottom=bottom,
            label=label,
            pattern_set_id=pattern_set_id,
            pattern_id=pattern_id
        )
        models.db.session.add(newann)
        models.db.session.commit()

        return newann.id

    # Deletes the annotation with the given ID, after some checks. Returns true
    # or false to indicate success.
    def deleteAnnotation(self, user_id, id):

        # Get the annotation in question
        annotationToDelete = models.Annotation.query.filter_by(id=id).first()

        # Verify the user has requested to delete his or her own annotation.
        if annotationToDelete.user_id != user_id:
            logging.critical(
                f"Error/securityissue on annotation deletion: User (id {user_id}) tried to delete an annotation (id {id}) belonging to another user (id {annotationToDelete.user_id}).")
            return False

        # Having reached this point, delete the annotation
        models.db.session.delete(annotationToDelete)
        models.db.session.commit()

        # Return true to indicate success
        return True

    # TODO(gus): Turn this into DataFrame output in line with Project.detectPatterns.
    def detectPatterns(self, type, series, thresholdlow=None, thresholdhigh=None, duration=300, persistence=.7, maxgap=300, series2=None, expected_frequency=0, min_density=0):

        # Determine the mode (see generateThresholdAlerts function description
        # for details on this parameter).
        if thresholdlow is None and thresholdhigh is None:
            # We must have at least one of thresholdlow or thresholdhigh.
            print("Error: We need at least one of threshold low or threshold high in order to perform pattern detection.")
            return []

        if thresholdhigh is None:
            mode = 0
            thresholdhigh = 0
        elif thresholdlow is None:
            mode = 1
            thresholdlow = 0
        else:
            mode = 2

        if type == 'patterndetection':

            # Find the series & run pattern detection
            for s in self.series:
                if s.id == series:
                    return s.generateThresholdAlerts(thresholdlow, thresholdhigh, mode, duration, persistence, maxgap, expected_frequency=expected_frequency, min_density=min_density).tolist()

        elif type == 'correlation':

            # TODO(Gus): TEMP!
            series2 = '/data/numerics/HR.HR:value'
            timewindow='10min'

            # The series2 parameter is required for correlation detection.
            if series2 is None:
                print("Error: Correlation detection requires series2 parameter.")
                return []

            # Find the series
            for s in self.series:
                if s.id == series:
                    s1 = s
                if s.id == series2:
                    s2 = s

            t1, v1 = s1.pullRawDataIntoMemory(returnValuesOnly=True)
            t2, v2 = s2.pullRawDataIntoMemory(returnValuesOnly=True)

            s1Data = pd.DataFrame({
                'time': t1,
                'val1': v1
            })
            s2Data = pd.DataFrame({
                'time': t2,
                'val2': v2
            })

            # Assemble into series 1 & series 2 dataframes
            # s1Data = pd.concat((t1, v1), axis=1)
            # s2Data = pd.concat((t2, v2), axis=1)
            #
            # # Assert column names
            # s1Data.columns = ['time', 'val1']
            # s2Data.columns = ['time', 'val2']

            # Create an inner joined dataset with common time column
            df = s1Data.merge(s2Data, on='time')

            # Filter out null values
            df = df[df.notnull()]

            # Create a datetime column
            df['time_conv'] = pd.to_datetime(df.time, unit='s')

            # Set the index to the datetime column
            df.set_index('time_conv', inplace=True)

            logging.debug(df)

            # Generate the rolling-window correlation
            corr = df['val1'].rolling(timewindow).corr(other=df['val2'])

            # Now run pattern detection on the rolling-window correlation
            alerts = generateThresholdAlerts(df['time'].to_numpy(), corr.to_numpy(), thresholdlow, thresholdhigh, mode, duration, persistence, maxgap).tolist()

            return alerts

        # Having reached this point, we were unable to generate the alerts.
        return []

    # Returns all event series
    def getEvents(self):

        logging.info(f"Assembling all event series for file {self.origFilePathObj}.")
        start = time.time()

        events = {}

        # If we're in realtime-mode, we have no events to return, so return now.
        if self.mode() == 'realtime':
            return events

        def catcols(row, idcol=None):
            idx = row[idcol]
            return (idx, '\n'.join(
                [f'{row[c]}' for c in row.index \
                    if (c is not idcol) and (row[c] is not None) and (row[c] != '')]))

        try:

            # Prepare references to HDF5 data
            meds = self.f['ehr/medications'][:]
            events['meds'] = meds.apply(catcols, 1, idcol='time').tolist()

        except Exception:
            print("Error retrieving meds.")
            traceback.print_exc()

        try:

            # Prepare references to HDF5 data
            ce = self.f['low_rate'][:]
            events['ce'] = ce.apply(catcols, 1, idcol='date').tolist()

        except Exception:
            print("Error retrieving ce.")
            traceback.print_exc()

        end = time.time()
        logging.info(f"Completed assembly of all event series for file {self.origFilePathObj}. Took {str(round(end - start, 5))}s.")

        return events

    # Produces JSON output for all series in the file at the maximum time range.
    def getInitialPayload(self, user_id):

        logging.info(f"Assembling all series full output for file {self.origFilePathObj}.")
        start = time.time()

        # Assemble the output object.
        outputObject = {
            'filename': self.origFilePathObj.name,
            'annotationsets': [{
                'id': 'general',
                'name': 'General',
                'description': None,
                'annotations': [annotationOrPatternOutput(a) for a in models.Annotation.query.filter_by(user_id=user_id, project_id=self.projparent.id, file_id=self.id, pattern_set_id=None).all()],
                'show': True,
            }] + [
                {
                    'id': patternset.id,
                    'name': patternset.name,
                    'description': patternset.description,
                    'annotations': [annotationOrPatternOutput(a) for a in models.Annotation.query.filter_by(user_id=user_id, project_id=self.projparent.id, file_id=self.id, pattern_set_id=patternset.id).all()],
                    'show': patternset.show_by_default,
                } for patternset in models.PatternSet.query.filter(models.PatternSet.project_id==self.projparent.id, or_(
                    models.PatternSet.id.notin_(
                        models.db.session.query(distinct(models.patternSetAssignments.c.pattern_set_id)).subquery()
                    ),
                    models.PatternSet.id.in_(
                        models.db.session.query(models.patternSetAssignments.c.pattern_set_id).filter(models.patternSetAssignments.c.user_id==user_id).subquery()
                    )
                )).all()
            ],
            'patternsets': [
                {
                    'id': patternset.id,
                    'name': patternset.name,
                    'description': patternset.description,
                    'patterns': [annotationOrPatternOutput(pattern) for pattern in models.Pattern.query.filter_by(pattern_set_id=patternset.id, file_id=self.id).all()],
                    'show': patternset.show_by_default,
                } for patternset in models.PatternSet.query.filter(models.PatternSet.project_id==self.projparent.id, or_(
                    models.PatternSet.id.notin_(
                        models.db.session.query(distinct(models.patternSetAssignments.c.pattern_set_id)).subquery()
                    ),
                    models.PatternSet.id.in_(
                        models.db.session.query(models.patternSetAssignments.c.pattern_set_id).filter(models.patternSetAssignments.c.user_id==user_id).subquery()
                    )
                )).all()
            ],
            'baseTime': 0, # ATW: Not sure if this is still necessary.
            'events': self.getEvents(),
            'metadata': self.getMetadata(),
            'series': {}
        }

        for s in self.series:
            outputObject['series'][s.id] = s.getFullOutput()

        end = time.time()
        logging.info(f"Completed assembly of all series full output for file {self.origFilePathObj}. Took {str(round(end - start, 5))}s.")

        # Return the output object
        return outputObject

    # Returns a dict of file metadata.
    def getMetadata(self):
        try:
            return self.f.file_meta
        except:
            return {}

    # Returns the series instance corresponding to the provided series ID, or
    # None if the series cannot be found.
    def getSeries(self, seriesid):

        # TODO(gus): We temporarily have to traverse both nueric and waveform
        # series until this is consolidated.
        for s in self.series:
            if s.id == seriesid:
                return s
        return None

    # Returns a list of series names available in the file.
    def getSeriesNames(self):

        seriesNames = []
        for s in self.series:
            seriesNames.append(s.id)
        return seriesNames

    # Retreves or creates & returns the series corresponding to the provided
    # series ID.
    def getSeriesOrCreate(self, seriesid):

        # Attempt to retrieve the series
        series = self.getSeries(seriesid)

        # Otherwise, create the series
        if series is None:

            # Create new series
            self.series.append(Series([seriesid], 'time', 'value', self))

            # Retrieve the newly-created series
            series = self.getSeries(seriesid)

        return series

    # Produces JSON output for a given list of series in the file at a specified
    # time range.
    def getSeriesRangedOutput(self, seriesids, start, stop):

        logging.info(f"Assembling series ranged output for file {self.origFilePathObj}, series [{', '.join(seriesids)}].")
        st = time.time()

        outputObject = {
            'series': {}
        }

        for s in self.series:
            if s.id in seriesids:
                outputObject['series'][s.id] = s.getRangedOutput(start, stop)

        et = time.time()
        logging.info(f"Completed assembly of series ranged output for file {self.origFilePathObj}, series [{', '.join(seriesids)}]. Took {str(round(et - st, 5))}s.")

        # Return the output object
        return outputObject

    # Loads the necessary data into memory for an already-processed data file
    # (does not load data though).Sets up classes for all series from file but
    # does not load series data into memory).
    def load(self):

        logging.info('Loading series from file.')

        # Iteratee through all datasets in the partial file
        for (ds, _) in self.f.recurse():
            self.loadSeriesFromDataset(ds)

        logging.info('Completed loading series from file.')

    # Load all available series from a dataset
    def loadSeriesFromDataset(self, ds):

        # Grab the columns.
        cols = ds.columns

        # Determine the time column name. Look for a column names 'time', 'timestampe',
        # or the first time column, in that order.
        timecol = None
        if 'time' in cols:
            timecol = 'time'
        elif 'timestamp' in cols:
            timecol = 'timestamp'
        else:
            for col in cols:
                if cols[col]['type'] == 'time':
                    timecol = col
                    break

        # If time column not available, print an error & return
        if timecol is None:
            logging.error(f"Did not find a time column in {self.origFilePathObj}.\nCols: {cols}")
            return
        else:
            logging.info(f'Using time column {ds.name}:{timecol}')

        # Iterate through all remaining columns and instantiate series for each
        for valcol in (c for c in cols if c != timecol):
            coltype = cols[valcol]['type']
            if coltype in ('real', 'integer'):
                logging.info(f'  - Adding {coltype} series: {ds.name}:{valcol}')
                self.series.append(Series(ds, timecol, valcol, self))
            else:
                logging.warning(f'  - Skipping unsupported {coltype} series: {valcol}')

    # Returns the mode in which File is operating, either "file" or "realtime".
    def mode(self):
        # TODO(gus): When reviving realtime functionality, revise this
        return 'file'

    # Process and store all downsamples for all series for the file.
    def process(self):

        try:

            logging.info(f"Processing & storing all series for file {self.origFilePathObj}.")
            start = time.time()

            # Print user message
            print(f"Downsampling file {self.name}...")

            # Create the file for storing processed data.
            self.pf = audata.File.new(str(self.procFilePathObj), overwrite=False, time_reference=self.f.time_reference, return_datetimes=False)

            # Process & store numeric series
            for s in self.series:
                s.processAndStore()

            self.pf.flush()

            # Reload series data in memory since we have new downsamples
            self.load()

            # Print user message
            print("Done.")

            end = time.time()
            logging.info(f"Completed processing & storing all series for file {self.origFilePathObj}. Took {str(round((end - start) / 60, 3))} minutes).")

        except (KeyboardInterrupt, SystemExit):

            logging.warning("Interrupt detected. Aborting as requested.")
            logging.warning(f"Deleting partially completed processed file {self.procFilePathObj}.")

            # Close the file
            try:
                self.pf.close()
            except Exception as e:
                logging.error(f"Unable to close the processed file.\n{e}\n{traceback.format_exc()}")

            # Delete the processed data file
            self.procFilePathObj.unlink()

            logging.info("Partial processed file has been removed.")

            # Quit the program
            quit()

        except Exception as e:

            logging.error(f"There was an exception while processing & storing data for {self.origFilePathObj}.\n{e}\n{traceback.format_exc()}\nDeleting partially completed processed file {self.procFilePathObj}.")

            # Close the file
            try:
                self.pf.close()
            except Exception as e:
                logging.error(f"Unable to close the processed file.\n{e}\n{traceback.format_exc()}")

            # Delete the processed data file
            self.procFilePathObj.unlink(missing_ok=True)

            logging.info("File has been removed.")

            # Re-raise the exception
            raise

    # Update an annotation with new values
    def updateAnnotation(self, user_id, id, left=None, right=None, top=None, bottom=None, seriesID='', label=''):

        # Get the annotation in question
        annotationToUpdate = models.Annotation.query.filter_by(id=id).first()

        # Verify the user has requested to update his or her own annotation.
        if annotationToUpdate.user_id != user_id:
            logging.critical(f"Error/securityissue on annotation update: User (id {user_id}) tried to update an annotation (id {id}) belonging to another user (id {annotationToUpdate.user_id}).")
            return False

        # Set updated values
        annotationToUpdate.series = seriesID
        annotationToUpdate.left = left
        annotationToUpdate.right = right
        annotationToUpdate.top = top
        annotationToUpdate.bottom = bottom
        annotationToUpdate.label = label

        # Commit the changes
        models.db.session.commit()

        # Return true to indicate success
        return True


"""Base element class."""
from typing import Optional, Union, Dict, Any
import h5py as h5

from audata._utils import json2dict, dict2json


class Element:
    """
    Represents an abstract `audata` element (e.g., files, groups, etc..) It should not
    be necessary to interact with this class directly.
    """

    def __init__(self,
                 parent: Optional[Union[h5.File, 'Element']] = None,
                 name: str = ''):
        if parent is None:
            self.parent = None
            self.file = None
            self._h5 = None
        elif isinstance(parent, h5.File):
            self.parent = self
            self.file = self
            self._h5 = parent
        elif isinstance(parent, Element):
            self.parent = parent
            self.file = parent.file
            self._h5 = parent._h5 if name == '' else parent._h5[name]
        else:
            raise Exception(f'Invalid parent: {type(parent)}.')

    def clear(self):
        """Clear members to reset class."""
        self.parent = None
        self.file = None
        self._h5 = None

    @property
    def hdf(self) -> Optional[h5.HLObject]:
        """Get wrapped HDF object."""
        return self._h5

    @property
    def name(self) -> Optional[str]:
        """Element name (`Optional[str]`, read-only)"""
        return self._h5.name if self._h5 is not None else None

    @property
    def filename(self) -> Optional[str]:
        """Path to file (`Optional[str]`, read-only)"""
        return self._h5.file.filename if self._h5 is not None else None

    @property
    def time_reference(self) -> Optional['dt.datetime']:
        """File time reference (`Optional[dt.datetime]`, read-only)"""
        return self.file.time_reference if self.file is not None else None

    @property
    def valid(self) -> bool:
        """Is element valid? (`bool`, read-only)"""
        return \
            self.parent is not None and \
            self.file is not None and \
            self._h5 is not None

    @property
    def meta(self) -> Dict[str, Any]:
        """Element meta data (HDF5 .meta attribute) (JSON dictionary)"""
        return json2dict(self._h5.attrs['.meta']) if self.valid and '.meta' in self._h5.attrs else {}

    @meta.setter
    def meta(self, data: Dict[str, Any]):
        if not self.valid:
            raise Exception('Attempting to set meta on invalid element!')
        self._h5.attrs['.meta'] = dict2json(data)

    @property
    def file_meta(self) -> Dict[str, Any]:
        """File metadata (HDF5 .meta attribute of the built-in root group) (JSON dictionary, read-only)"""

        # Get the .meta attribute if it exists, otherwise return empty object now
        try:
            ms = self._h5.file.attrs['.meta']
        except:
            print("File metadata not found.")
            return {}

        # Parse and return the .meta json object
        try:
            return json2dict(ms)
        except Exception as e:
            print(f"There was an error decoding the file metadata!\n\nMetadata JSON:\n{self._h5.file.attrs['.meta'][0]}\n\nJSON decode exception:\n{e}\n")
            return {}

    @file_meta.setter
    def file_meta(self, data: Dict[str, Any]):
        if not self.valid:
            raise Exception('Attempting to set file meta on invalid elemenet!')
        self._h5.file.attrs['.meta'] = dict2json(data)



"""Script to generate some sample data."""
import os
import datetime as dt

import pandas as pd
import numpy as np
import lorem
import tzlocal

if __name__ == '__main__':

    def ensure(filename):
        """Ensure directory to store file exists."""
        filename = 'patient123/{}'.format(filename)
        os.makedirs(os.path.dirname(filename), exist_ok=True)
        return filename

    time_reference = dt.datetime.now(tzlocal.get_localzone())
    PERIOD = 60 * 30  # s

    print('Making low-rate...')
    SAMPLE_FREQ = 1. / 5  # Hz
    SAMPLES = int(SAMPLE_FREQ * PERIOD)
    pd.DataFrame(
        data={
            'time': [
                time_reference + dt.timedelta(seconds=x / SAMPLE_FREQ)
                for x in range(SAMPLES)
            ],
            'time2': [
                time_reference + dt.timedelta(seconds=x / SAMPLE_FREQ + 0.5)
                for x in range(SAMPLES)
            ],
            'ints':
                list(range(SAMPLES)),
            'floats': [i * 1.1 for i in list(range(SAMPLES))],
            'strings': [lorem.sentence() for x in range(SAMPLES)],
            'factor':
                pd.Series([['Cat', 'Dog', 'Liger'][x % 3]
                           for x in range(SAMPLES)],
                          dtype='category')
        }).to_csv(ensure('some_table.csv'), index=False)

    print('Making 500Hz ECG...')
    SAMPLE_FREQ = 500  # Hz
    SAMPLES = int(SAMPLE_FREQ * PERIOD)
    pd.DataFrame(
        data={
            'time': [x / SAMPLE_FREQ for x in range(SAMPLES)],
            'II': np.random.uniform(-2, 2, SAMPLES),
            'III': np.random.uniform(-2, 2, SAMPLES),
            'IV': np.random.uniform(-2, 2, SAMPLES)
        }).to_csv(ensure('waveform/ecg.csv'), index=False)

    print('Making 250Hz waveform...')
    SAMPLE_FREQ = 250  # Hz
    SAMPLES = int(SAMPLE_FREQ * PERIOD)
    pd.DataFrame(
        data={
            'time': [x / SAMPLE_FREQ for x in range(SAMPLES)],
            'ART': np.random.uniform(50, 90, SAMPLES),
            'PAP': np.random.uniform(10, 15, SAMPLES),
            'CVP': np.random.uniform(0, 5, SAMPLES)
        }).to_csv(ensure('waveform/pressures.csv'), index=False)

    print('Making vitals...')
    SAMPLE_FREQ = 1.5  # Hz
    SAMPLES = int(SAMPLE_FREQ * PERIOD)
    pd.DataFrame(
        data={
            'time': [x / SAMPLE_FREQ for x in range(SAMPLES)],
            'Sys': np.random.uniform(80, 90, SAMPLES)
        }).to_csv(ensure('vitals/some_device/Sys.csv'), index=False)
    pd.DataFrame(
        data={
            'time': [x / SAMPLE_FREQ for x in range(SAMPLES)],
            'MAP': np.random.uniform(70, 80, SAMPLES)
        }).to_csv(ensure('vitals/some_device/MAP.csv'), index=False)
    pd.DataFrame(
        data={
            'time': [x / SAMPLE_FREQ for x in range(SAMPLES)],
            'Dia': np.random.uniform(50, 65, SAMPLES)
        }).to_csv(ensure('vitals/some_device/Dia.csv'), index=False)
